<?php

namespace Sale\Handlers\PaySystem;

use Bitrix\Main\Error;
use Bitrix\Main\Loader;
use Bitrix\Main\Context;
use Bitrix\Main\Request;
use Bitrix\Main\SystemException;
use Bitrix\Sale\Order;
use Bitrix\Sale\Payment;
use Bitrix\Sale\PaySystem\ServiceHandler;
use Bitrix\Sale\PaySystem\ServiceResult;
use Bitrix\Sale\PaySystem\IRefund;
use Vtb\Pay\VtbHandler;
use Vtb\Pay\VtbPayLogger;
use Vtb\Pay\VtbPayException;

error_reporting(E_ERROR | E_WARNING | E_PARSE);

if (!Loader::includeModule('vtbpayments.ecom')) {
    throw new SystemException('vtbpayments.ecom module not found');
}

class vtbpaymentsHandler extends ServiceHandler implements IRefund
{
    // Объект заказа
    private $order;

    // Объект логирования
    private $logger;

    // Настройки модуля
    private array $params;

    // ЭтоВозврат
    private bool $isRefund = true;

    /**
     * @param Payment $payment
     * @param Request|null $request
     */
    public function initiatePay(Payment $payment, Request $request = null)
    {
        $this->order = Order::load($payment->getOrderId());

        if (empty($this->params)) {
            $this->params = $this->getPaymentParamas($payment);
        }

        $this->setVtbpayLogger();

        if (empty($this->params['VTB_CLIENT_ID']) || empty($this->params['VTB_CLIENT_SECRET'])) {
            // Error, client do not configure payment module
            $this->logger->error('Module not configured');
            require __DIR__ . '/configure.php';
        }

        $orderId = $payment->getOrderId();

        if ($this->params['VTB_TEST_MODE'] === 'Y') VtbHandler::setTest();

        $contextRequest = Context::getCurrent()->getRequest();
        $queryData = $contextRequest->getQueryList()->toArray(); // массив get параметров

        $this->logger->setOption('additionalCommonText', 'init-' . rand(1111, 9999));
        $this->logger->debug('init-data:', [
            'query' => $queryData,
            'paid' => $payment->getField('PAID')
        ]);

        $orderStatus = [];
        if ($payment->getField('PAID') !== 'N' || !empty($queryData['orderId'])) {

            $this->logger->setOption('additionalCommonText', 'return-' . rand(1111, 9999));

            $orderStatus = $this->getOrderStatus($payment, $orderId);
        }

        // Checks
        if ($orderStatus && $orderStatus['object']['status']['value'] === 'PAID' && ($payment->getField('PAID') === 'N' && !isset($statusOrder['object']['transactions']['refunds']))) {

            $this->setTransaction($payment, 'Y', 'P');

            if ($payment->isPaid()) {
                $this->logger->debug("Order {$payment->getField('ORDER_ID')} paid");
                LocalRedirect(
                    $this->getResultUrl($payment->getField('ORDER_ID')),
                    true
                );
            }

        } else if ($orderStatus && isset($orderStatus['error']) && $payment->getField('PAID') === 'N') {

            $this->checkPayGatewayAnswer($orderStatus, $orderId);

        } else if ($orderStatus && $orderStatus['object']['status']['value'] === 'CREATED') {

            $this->logger->debug('Payment uncomplete');
            // Error, transaction do not pay
            return $this->getPayState($payment, $orderStatus['object'], $orderId);

        } else if ($orderStatus && $orderStatus['object']['status']['value'] === 'PENDING') {

            $this->setTransaction($payment, 'Y', 'VP');

            if ($payment->isPaid()) {
                $this->logger->debug("Order {$payment->getField('ORDER_ID')} pending");
                $url = $this->getResultUrl($payment->getField('ORDER_ID'));
                LocalRedirect($url, true);
            }

        } else {

            if ($payment->getField('PAID') === 'Y' &&
            $payment->getField('STATUS_ID') === $this->params['VTB_ORDER_STATUS']) {

                $this->logger->debug("Payment {$orderId} complete");
                // Local redirect to SUCCESS page
                LocalRedirect($this->getResultUrl($orderId), true);
            }

        }
        return $this->getPayState($payment);
    }

    /**
     * @param Payment $payment
     * @param Request $request
     */
    public function processRequest(Payment $payment, Request $request)
    {
        sleep(5);
        $this->order = Order::load($payment->getOrderId());

        if (empty($this->params)) {
            $this->params = $this->getPaymentParamas($payment);
        }

        $this->setVtbpayLogger();
        $this->logger->setOption('additionalCommonText', 'webhook-' . rand(1111, 9999));

        // Receive incoming data
        $json = json_decode($request->getInput(), true);
        $opid = explode('-', $json['object']['orderId']) ?? ['0', '0'];

        $this->logger->debug('callback-data:', $json);

        // explode orderId
        $orderId = $opid[0];
        $paymentId = $opid[1];

        if ($orderId + $paymentId <= 0) { // Callback is empty or broken
            $this->logger->error("Request for order: {$json['object']['orderId']} is fail");
            return false; // exit
        }

        $statusOrder = $this->getOrderStatus($payment, $orderId);
        $paymentStatus = $statusOrder['object']['status']['value'] ?? '';


        if (
            $paymentStatus !== 'REFUNDED' &&
            isset($statusOrder['object']['transactions']['refunds'])
        ) {
            // Указываем статус возврата
            if ($this->order->getField('STATUS_ID') !== 'VR') {
                $this->setTransaction($payment, 'N', 'VR', 'Y');
            }
        }
        elseif ($paymentStatus === 'PAID' && !isset($statusOrder['object']['transactions']['refunds'])) {
            if ($this->order->getField('STATUS_ID') !== 'P' && ($this->order->getField('STATUS_ID') !== 'VR' || $this->order->getField('STATUS_ID') !== 'VC')) {
                $this->setTransaction($payment, 'Y', 'P');
            }
        }
        elseif ($paymentStatus === 'PENDING') {
            if ($this->order->getField('STATUS_ID') !== 'VP') {
                $this->setTransaction($payment, 'Y', 'VP');
            }
        }
        elseif ($paymentStatus === 'REFUNDED') {
            if ($this->order->getField('STATUS_ID') !== 'VR') {
                $this->setTransaction($payment, 'N', 'VR', 'Y');
            }
        }
        else {
            if ($this->order->getField('STATUS_ID') !== 'VC') {
                $this->setTransaction($payment, 'N', 'VC', 'Y');
            }
        }

        return 'ok';
    }

    /**
     * @param Request $request
     * @return false|mixed|string
     */
    public function getPaymentIdFromRequest(Request $request)
    {
        $json = json_decode($request->getInput(), true);
        $paymentId = explode('-', $json['object']['orderId'])[1]; // split order & payment ids

        if ($paymentId <= 0) {
            return false;
        }

        return $paymentId;
    }

    /**
     * @param Request $request
     * @param $paySystemId
     * @return true
     */
    public static function isMyResponse(Request $request, $paySystemId)
    {
        // Skip
        return true;
    }

    /**
     * @param $orderStatus
     * @param $orderId
     * @return void
     */
    private function checkPayGatewayAnswer($orderStatus, $orderId)
    {
        // Error transaction, do not change order status
        if ($orderStatus['error']['id'] == '10000007') {
            // Pay gateway return error
            $this->logger->error("Order {$orderId} take error code: {$orderStatus['error']['id']}");

            // Local redirect to FAIL page
            LocalRedirect($this->getResultUrl($orderId), true);
        } else {
            // Order do not paid
            $this->logger->error("Order {$orderId} not paid");
        }
    }


    private function setTransaction($payment, $isPaid, $statusId, $isCanceled = null)
    {
        $this->logger->debug("Status for order {$payment->getOrderId()} change to {$statusId}");

        // Complete transaction, change payment status and mark order to paid
        $payment->setPaid($isPaid);
        $payment->setField('PAID', $isPaid);
        if ($isCanceled == 'Y') $payment->getOrder()->setField('CANCELED', $isCanceled);
        $payment->getOrder()->setField('STATUS_ID', $statusId);
        $payment->getOrder()->save();
    }

    /**
     * @return string[]
     */
    public function getCurrencyList(): array
    {
        return [
            'RUB'
        ];
    }

    /**
     * @param Payment $payment
     * @param int $orderId
     * @return array|false
     */
    private function getOrderStatus($payment, $orderId = 0)
    {
        if (empty($orderId)) {
            $orderId = $payment->getField('ORDER_ID');
        }

        if (empty($this->params)) {
            $this->params = $this->getPaymentParamas($payment);
        }

        if ($this->params['VTB_TEST_MODE'] === 'Y') VtbHandler::setTest();

        if ($order = Order::load($orderId)) {
            return VtbHandler::getOrderInfo(
                $order,
                $payment->getField('ID'),
                $this->params,
                $orderId
            );
        }
        return false;
    }

    /**
     * @param $payment
     * @param $orderStatus
     * @param $orderId
     */
    private function getPayState($payment, $orderStatus = 'NEW', $orderId = 0)
    {
        if (empty($this->params)) {
            $this->params = $this->getPaymentParamas($payment);
        }

        if (empty($orderId)) {
            $orderId = $payment->getOrderId();
        }

        $order = Order::load($orderId);

        $orderAmount = CurrencyFormat($payment->getSum(), $this->getCurrencyList()[0]);

        $this->logger->setOption('additionalCommonText', 'payment-' . rand(1111, 9999));
        $url = '';

        if ($order && $orderStatus === 'NEW') {

            $VtbHandler = new VtbHandler();

            $url = $VtbHandler->getUrl(
                $order,
                $this->params,
                $payment,
                $orderId
            );

            if (isset($url['payUrl'])) {
                $url = $url['object']['payUrl'];
            }

        } else if ($order && $orderStatus['status']['value'] === 'CREATED') {
            $url = $orderStatus['payUrl'];
        }

        $args = [];
        if (!empty($url)) {
            $urlArray = explode('?', $url);
            parse_str($urlArray[1], $urlParams);
            $args = [
                'URL' => $url,
                'ACTION_URL' => $urlArray[0],
                'PARAMS' => $urlParams,
                'AMOUNT' => $orderAmount,
                'ERROR' => 'OK'
            ];
        } else {
            $args['ERROR'] = 'REQUEST';
        }

        $args['WINDOW'] = false;
        if ($this->params['VTB_NEW_WINDOW'] === 'Y') {
            $args['WINDOW'] = true;
        }
        $this->setExtraParams($args);

        return $this->showTemplate($payment, 'template');
    }

    /**
     * @param Payment $payment
     * @return array
     */
    private function getPaymentParamas(Payment $payment): array
    {
        if (empty($this->params)) {
            $this->params = $this->getParamsBusValue($payment);
        }
        $service = $this->service->getFieldsValues();

        return [
            // GLOBAL DEBUGGING MODE
            'VTB_DEBUG_MODE' => $this->params['VTB_DEBUG_MODE'],
            // PAYMENT
            'VTB_CLIENT_ID' => $this->params['VTB_CLIENT_ID'],
            'VTB_CLIENT_SECRET' => $this->params['VTB_CLIENT_SECRET'],
            'VTB_MERCHANT_AUTHORIZATION' => $this->params['VTB_MERCHANT_AUTHORIZATION'],
            'VTB_TEST_MODE' => $this->params['VTB_TEST_MODE'],
            'VTB_ORDER_STATUS' => $this->params['VTB_ORDER_STATUS'],
            'VTB_STEP_STAGE' => $this->params['VTB_STEP_STAGE'],
            'VTB_NEW_WINDOW' => $service['NEW_WINDOW'],
            // FISCAL
            'VTB_FISCAL_ENABLE' => $this->params['VTB_FISCAL_ENABLE'], // 1.4.3 - Добавлена настройка вкл/выкл фискализации
            'VTB_FISCAL_EMAIL' => $this->params['VTB_FISCAL_EMAIL'],
            'VTB_FISCAL_TAX' => $this->params['VTB_FISCAL_TAX'],
            'VTB_FISCAL_DELIVERY' => $this->params['VTB_FISCAL_DELIVERY'],
            'VTB_FISCAL_MEASURE' => $this->params['VTB_FISCAL_MEASURE'],
            'VTB_FISCAL_PAYMENT_TYPE' => $this->params['VTB_FISCAL_PAYMENT_TYPE'],
            'VTB_FISCAL_SUBJECT' => $this->params['VTB_FISCAL_SUBJECT']
        ];
    }

    /**
     * @param string $orderId
     * @return array|string|string[]
     */
    private function getResultUrl(string $orderId)
    {

        $url = $_SERVER['REQUEST_URI'];

        if (count(explode('-', $orderId)) > 1) {
            $order = explode('-', $orderId)[0]; // Set orderId

            $url = preg_replace('/\d+\-\d+/iu', $order, $url);
        }

        // Replace code to orderId
        $url = '/personal/orders/'.$orderId; // 1.3.1 - теперь адрес возврата жестко прописан

        return $url;
    }


    /**
     * 1.4.1 - добавлена возможность оформления возврата при оплате картой
     * Отправляем запрос на возврат в ВТБ, возвращвем результат запроса и логируем входящие и выходящие данные.
     * Если запрос прошёл успешно, то и в CMS отображаем информацию о возврате.
     *
     * @return object
     */
    public function refund(Payment $payment, $refundableSum): ServiceResult
    {
        $this->setVtbpayLogger();
        $this->logger->setOption('additionalCommonText', 'refund-' . rand(1111, 9999));

        $result = new ServiceResult();
        $order = Order::load($payment->getOrderId());
        if (empty($this->params)) {
            $this->params = $this->getPaymentParamas($payment);
        }
        if ($this->params['VTB_TEST_MODE'] === 'Y') VtbHandler::setTest();

        $orderId = $payment->getField('ORDER_ID');
        try {
            VtbHandler::setPaymentRefunds($this->params, $orderId, $payment->getField('ID'));

            // Под большим вопросом, нужен ли код ниже для успешного возврата
            $result->setOperationType(ServiceResult::MONEY_LEAVING);
            if ($order->getField('STATUS_ID') != 'VR') {
                $this->setTransaction($payment, 'N', 'VR', 'Y');
                $payment->setPaid('N');
                $order->setFieldNoDemand('PAYED', 'N');
                $return = $payment->setReturn('P');
                if (!$return->isSuccess())
                {
                    $order->setField('COMMENTS', $return->getErrorMessages());
                    $result->addError(new Error($return->getErrorMessages())); // 1.4.2 - добавлен вывод ошибки
                }
            }
        } catch (\Exception | VtbPayException $e) {
            $order->setField('COMMENTS', $e->getMessage());
            $result->addError(new Error($e->getMessage()));
        }

        $order->save();
        return $result;
    }

    /**
     * Инициализация и настройка объекта класса VtbPayLogger.
     *
     * Эта функция инициализирует и настраивает логгер, используемый плагином VtbPay для ведения журнала.
     *
     * @return void
     */
    private function setVtbpayLogger(): void
    {
        $logger = VtbPayLogger::getInstance();
        $this->logger = $logger
            ->setOption('showBacktrace', true)
            ->setOption('maskSensitiveData', true)
            ->setLogFilePath(dirname(__DIR__, 1) . '/logs/vtbpay-' . date('d-m-Y') . '.log')
            ->setCustomRecording(function($message) use ($logger) {
                if (
                    isset($this->params['VTB_DEBUG_MODE']) &&
                    $this->params['VTB_DEBUG_MODE'] === 'Y'
                ) $logger->writeToFile($message);
            }, VtbPayLogger::LOG_LEVEL_DEBUG);
    }
}
